#include <kqml/kqmlconf.h>

#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <sys/ioctl.h>
#if !defined (FIONREAD)
#	include <sys/filio.h> /* Linux does not have filio.h */
#endif

#ifdef __KQML_MACH_SUN__
extern char *sys_errlist[];
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

/*
 * Exported Functions:
 * kqml_stricmp
 * kqml_strincmp
 * kqml_match_agent_names
 * kqml_digit_count
 * kqml_digit_to_str
 */

/*
 * kqml_stricmp 
 *
 * Like strcmp, but ignores the case while comparing the strings
 *
 * Returns 0 if the strings are same, -ve value if s1 is lexically less than
 * s2, +ve value if s1 is lexically greater than s2
 */
int 
kqml_stricmp (char *s1, char *s2)
{
	char c1, c2;

	while (1)
	{
		c1 = toupper (*(s1++));
		c2 = toupper (*(s2++));

		if (c1 != c2 || c1 == '\0')
		{
			return c1 - c2;
		}
	}
}

/*
 * kqml_strincmp 
 *
 * Like strncmp, but ignores the case while comparing the strings
 *
 * Returns 0 if the strings are same, -ve value if s1 is lexically less than
 * s2, +ve value if s1 is lexically greater than s2
 */
int 
kqml_strincmp (char *s1, char *s2, int n)
{
	char c1, c2;

	if (n <= 0)
		return 0;

	while (1)
	{
		c1 = toupper (*(s1++));
		c2 = toupper (*(s2++));

		if (c1 != c2 || c1 == '\0' || --n == 0)
		{
			return c1 - c2;
		}
	}
}

int 
kqml_digit_count (long number)
{
	int count;

	if (number < 0)
	{
		count = 1;
	}
	else
	{
		count = 0;
	}

	do
	{
		count++;
		number /= 10;
	}
	while (number != 0);

	return count;
}

int 
kqml_digit_to_str (long number, char *str)
{
	int count, i, endI;
	long rem;

	count = kqml_digit_count (number);

	if (number < 0)
	{
		endI = 1;
		str[0] = '-';
	}
	else
	{
		endI = 0;
	}

	for (i = count-1; i >= endI; i--)
	{
		rem = number % 10;
		number /= 10;
		str[i] = (rem < 0 ? -rem : rem) + '0';
	}

	return count;
}

int 
kqml_Udigit_count (unsigned long number)
{
	int count;

	count = 0;
	do
	{
		count++;
		number /= 10;
	}
	while (number != 0);

	return count;
}

int 
kqml_Udigit_to_str (unsigned long number, char *str)
{
	int count, i;
	unsigned long rem;

	count = kqml_Udigit_count (number);

	for (i = count-1; i >= 0; i--)
	{
		rem = number % 10;
		number /= 10;
		str[i] = rem + '0';
	}

	return count;
}

int 
_kqml_pending_bytes (int fd, FILE *fp)
{
	int size;

	if (ioctl(fd, FIONREAD, &size) <= -1)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("Could not do FIONREAD in Con #%d. %s.", fd, sys_errlist[errno]);
		size = 0;
	}

	if (fp)
	{
#		if defined (__KQML_MACH_LINUX__)
			size += fp->_IO_read_end - fp->_IO_read_ptr;
#		else
			size += fp->_cnt;
#		endif
	}

	return size;
}


/******	This subroutine matches the Fully Qualified Agent Names with 
			paritally qualified agent names  -- Anupama.  ******/

int kqml_match_agent_names(char *FQAN, char *name)
{
	char	*dupname, *substr;

	if ((dupname = strdup(FQAN)) == NULL)
	{
		fprintf (stderr, "Error : Unable to allocate space for strdup\n");
		exit(1);
	}
	
	/*** If name and FQAN are same return 0 ***/
	if (kqml_stricmp(name, FQAN) == 0) return 0;

	/*** While the names are not equal, find the next substring of FQAN
		  to match against the name of the agent ***/
	substr = rindex(dupname, '.');
	while (substr != NULL)
	{
		*substr = '\0';
		if (kqml_stricmp(name, dupname) == 0) return 0;
		substr = rindex(dupname, '.');
	}

	return -1;
}

