/*
 * $Log: send_msg.c,v $
 * Revision 1.1  1993/08/06  18:31:01  robin
 * Initial revision
 *
 *
 */

/*
 *	(c) Copyright 1993 Paramax Systems Corp.
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */


#include <kqml/kqmlconf.h>

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <time.h>
#include <unistd.h>

#ifdef __KQML_MACH_SVR4__
#	include <sys/systeminfo.h>	/* sysinfo instead of gethostname */
#	ifdef __KQML_MACH_SGI__
#		include <bstring.h>
#	endif
#endif

#include <kqml.h>
#include <kqml/kqmllib.h>

static int default_timeout = -1;

static void set_default_timeout ();
static char *gen_kmsg_id ();
static void set_kqml_address (char *receiver, kqml_message *msg);
static int check_for_reply (kqml_message *r_msg, kqml_message *s_msg);

static void
set_default_timeout ()
{
	char *temp;

	temp = getenv ("KQML_WAIT_TIME");
	if (temp)
	{
		char *dummy;

		default_timeout = strtol (temp, &dummy, 10);

		/* If the value is not a proper +ve no */
		if (*dummy != '\0' || default_timeout < 0)
		{
			default_timeout = 30;
		}
	}
	else
	{
		default_timeout = 30;
	}
}

static char *
gen_kmsg_id ()
{
	static int len = 0;
	static int kmsg_id_num = 0;
	static char this_agent_id[256 + 67];
	/* 256 characters for the hostname (internet names can be 256 chars),
		32 characters for pid, 32 characters for the id and 2 for the + and 
		one for the null  = 67*/

	if (len == 0)
	{
		sprintf (this_agent_id, "%lu+", getpid ());
		len = strlen (this_agent_id);

		/* Ignore errors. We cannot do anything about it */
#		ifdef __KQML_MACH_SVR4__
		sysinfo (SI_HOSTNAME, this_agent_id+len, 256);
#		else
		gethostname (this_agent_id+len, 256);
#		endif

		len = strlen (this_agent_id);
	}

	sprintf (this_agent_id+len, "+%d", kmsg_id_num++);

	return this_agent_id;
}

static void
set_kqml_address (char *receiver, kqml_message *msg)
{
	char *agent, *protocol, *addr_info;

	if (_kqml_ans_lookup (receiver, &agent, &protocol, &addr_info))
	{
		char *address;
		int len;

		len = strlen (agent) + strlen (protocol) + strlen (addr_info) + 5;
		address = malloc (len);
		if (address != NULL)
		{
			char *temp;

			temp = address;
			*temp++ = '(';
			temp = memccpy (temp, agent, 0, len);
			*(temp - 1) = ' ';
			temp = memccpy (temp, protocol, 0, len);
			*(temp - 1) = ' ';
			temp = memccpy (temp, addr_info, 0, len);
			*(temp - 1) = ')';
			*temp = '\0';

			_kqml_put_field_nd (":kqml-address", msg, address, len, 
				KQML_TYPE_SEXPR);
		}
	}
}

static int
check_for_reply (kqml_message *r_msg, kqml_message *s_msg)
{
	char *receiver, *reply_with, *in_reply_to, *sender;

	receiver = kqml_get_field (":receiver", s_msg, NULL, NULL);
	reply_with = kqml_get_field (":reply-with", s_msg, NULL, NULL);
	sender = kqml_get_field (":sender", r_msg, NULL, NULL);
	in_reply_to = kqml_get_field (":in-reply-to", r_msg, NULL, NULL);

	if (receiver && reply_with && sender && in_reply_to)
	{
		if (kqml_stricmp (reply_with, in_reply_to) == 0 &&
			(kqml_stricmp (sender, receiver) == 0 ||
			kqml_stricmp (sender, kqml_agent_name) == 0))
		{
			return TRUE;
		}
	}

	return FALSE;
}

int
kqml_send (int timeout_value, kqml_message *msg, kqml_message **reply)
{
	static char waiting_for_reply = FALSE;
	int blocking, res;
	time_t start_time, wait_time;
	kqml_message *reply_message;
	char *receiver;

	/* if timeout is not defined */
	if (default_timeout == -1)
	{
		set_default_timeout ();
	}

	if (timeout_value < 0)
	{
		timeout_value = default_timeout;
	}
	
	receiver = kqml_get_field (":receiver", msg, NULL, NULL);
	if (receiver == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) ("kqml_send: receiver not defined.");
		return -1;
	}

	if (_kqml_im_an_ans)
	{
		set_kqml_address (receiver, msg);
	}

	if (kqml_get_field (KMSG_ID_KEYWORD, msg, NULL, NULL) == NULL)
	{
		kqml_put_field (KMSG_ID_KEYWORD, msg, gen_kmsg_id (), -1);
	}

	/* If the :sender field is not defined, we will put the agent's name */
	if (kqml_get_field (":sender", msg, NULL, NULL) == NULL)
	{
		kqml_put_field (":sender", msg, kqml_agent_name, -1);
	}

	blocking = ! (_kqml_get_reply_with_field (msg) == NULL || reply == NULL);

	if (blocking)
	{
		if (! waiting_for_reply)
		{
			waiting_for_reply = TRUE;
		}
		else
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				("kqml_send: can block on only one request.");
			return -1;
		}
	}

	kqml_critical_section_start ();
	res = _kqml_send_xdr_msg (msg);

	if (!res)
	{
		kqml_critical_section_end ();
		waiting_for_reply = FALSE; 
		return -1; 
	}

	/* if nonblocking, you are done */
	if (!blocking)
	{
		if (reply)
		{
			*reply = NULL;
		}
		kqml_critical_section_end ();
		_kqml_dbg_msg (KQML_DBG_VERBOSE) 
			("kqml_send: non-blocking request. so return.");
		return 1;
	}

	_kqml_dbg_msg (KQML_DBG_VERBOSE) 
		("kqml_send: blocking request. so don't return yet.");
  
	reply_message = NULL;
	while (reply_message == NULL)
	{
		/* sleep until reply */
		start_time = time (NULL);
		res = _kqml_wait_for_msg (timeout_value);
		wait_time = time (NULL) - start_time;

		if (res == -1)
		{
			_kqml_dbg_msg (KQML_DBG_ALWAYS)
				("kqml_send: Wait error on blocked request.");
			break;
		}

		if (res != 0)
		{
			_kqml_dbg_msg (KQML_DBG_VERBOSE) ("Send Handling message from child.");

			do
			{
				reply_message = _kqml_fetch_xdr_msg ();

				if (reply_message == NULL)
				{
					_kqml_dbg_msg (KQML_DBG_VERBOSE) 
						("No KQML message. So, return already.");	
					break;
				}
				else
				{
					_kqml_dbg_msg (KQML_DBG_VERBOSE)
						("Waiting for reply. So check this message.");

					if (check_for_reply (reply_message, msg))
					{
						_kqml_dbg_msg (KQML_DBG_VERBOSE)
							("Don't call handler. This is the reply.");
						break;
					}
					else
					{
						_kqml_handler (reply_message);
						reply_message = NULL;
					}
				}
			}
			while (_kqml_msgs_in_queue ());
		}

		if (reply_message == NULL && timeout_value > 0)
		{
			timeout_value -= wait_time;
			if (timeout_value <= 0)
			{
				_kqml_dbg_msg (KQML_DBG_ALWAYS) 
					("kqml_send: Timeout waiting for reply.");
				break;
			}
		}
	}

	waiting_for_reply = FALSE;
	kqml_critical_section_end ();

	if (reply_message == NULL)
	{
		return -1;
	}

	if (reply)
	{
		*reply = reply_message;
	}

	/* if error return */
	if (kqml_stricmp (kqml_get_performative (reply_message), "error") == 0)
	{
		return 0;
	}
	else
	{
		return 1;
	}
}

/*
 * kqml_send_msg
 *
 * Send a message to a another agent for execution and return of results.
 *
 * Parameters:
 *
 *	1) performative
 *         The performative to be used for this KQML expression.
 *
 *	2) content
 *	   The query to be executed by the remote database.  This should
 *	   be a pointer to a character array.  The ending ';' must not be
 *	   included in the query string.
 *
 *	3) rec_agent
 *	   This is the symbolic name of the agent to which the message
 *	   is to be sent.  This argument must be a pointer to a character
 *         string.
 *
 *      4) reply
 *         This is a flag that will be 0 if no reply is requested or some
 *         non-zero number if a reply is requested.
 *
 * Value Returned:
 *
 *	Pointer to a character string - OK, the query has been successfully
 *		executed.
 *	NULL - Failure.  Either the query was unable to be executed on the
 *		remote database or there was a problem during transmission
 *		of the query to or from the database wrapper.
 *
 */

int
kqml_send_msg (int timeout_value, char *perf, char *content, char *rec_agent, 
	char *reply_with, kqml_message **reply, ...)
{
	char *field, **field_array, **value_array;
	va_list fp;
	int i, res, field_count;
	char end_of_arg;

	va_start (fp, reply);
	field_count = 0;
	end_of_arg = FALSE;
	while (! end_of_arg)
	{
		field = va_arg (fp, char *);
		if (field != NULL)
		{	
			field_count++;
			if (va_arg (fp, char *) == NULL)
			{
				_kqml_dbg_msg (KQML_DBG_VERBOSE) 
					("kqml_send_msg: field %s has NULL value. Ignored.", field);
				end_of_arg = TRUE;
			}
		}
		else
		{
			end_of_arg = TRUE;
		}
	}
	va_end (fp);

	field_array = malloc (sizeof (char *) * (field_count + 1));
	if (field_array == NULL)
	{
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_send_msg: Could not set field array.");
		return -1;
	}

	value_array = malloc (sizeof (char *) * (field_count + 1));
	if (value_array == NULL)
	{
		free (field_array);
		_kqml_dbg_msg (KQML_DBG_ALWAYS) 
			("kqml_send_msg: Could not set value array.");
		return -1;
	}

	i = 0;
	va_start (fp, reply);
	for (i = 0; i < field_count; i++)
	{
		field_array[i] = va_arg (fp, char *);
		value_array[i] = va_arg (fp, char *);
	}
	field_array[i] = NULL;
	value_array[i] = NULL;

	res = kqml_send_msgv (timeout_value, perf, content, rec_agent, reply_with, 
				reply, field_array, value_array);

	free (field_array);
	free (value_array);

	return res;
}

/*
 * kqml_send_msgv
 *
 * This function does the same work that kqml_send_msg, above, does, but 
 * can be used when applications do not know at compile time how many 
 * additional keyword/value pairs will be passed.
 *
 */

int
kqml_send_msgv (int timeout_value, char *perf, char *content, char *rec_agent, 
	char *reply_with, kqml_message **reply, char *field[], char *value[])
{
	kqml_message *omsg;
	int res;

	omsg = kqml_build_msgv (perf, content, rec_agent, reply_with, field, value);
	if (omsg == NULL)
	{
		return -1;
	}

	res = kqml_send (timeout_value, omsg, reply);

	kqml_free_msg (omsg);
	return res;
}
